import java.io.*;
import java.util.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * <p>Title: Naive Bayesian Classifier</p>
 * <p>Description: A naive bayesian classifier</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Department of Computer Science, Univ. of Illinois at Chicago</p>
 * @author Huiyong Xiao
 * @version 1.0
 */

public class MainDialog
    extends JDialog {
  // two variables for identifying whether the training or testing has been done.
  private boolean trained = false;
  private boolean tested = false;

  private File namefile = null;
  private File datafile = null;
  private File testfile = null;

  ArrayList classlist = new ArrayList(); // the list of classes
  ArrayList attrlist = new ArrayList(); // the list of attribute names
  ArrayList valuelist = new ArrayList(); // the list of values for each attribute
  int tupleNum = 0;

  Hashtable pTable = new Hashtable(); // the hashtable containing the probabilities.
  ArrayList pList = new ArrayList(); // just for printing the rule list

  int[][] confusionMatrix = null;
  double errorRate = 0;

  JPanel panel1 = new JPanel();
  GridBagLayout gridBagLayout1 = new GridBagLayout();
  JTextField fileNameFiled = new JTextField();
  JButton browseButton = new JButton();
  JLabel jLabel1 = new JLabel();
  JButton testButton = new JButton();
  JButton trainButton = new JButton();
  JButton ruleSaveButton = new JButton();
  JButton resultClearButton = new JButton();
  JButton resultSaveButton = new JButton();
  JScrollPane jScrollPane1 = new JScrollPane();
  JTextArea ruleArea = new JTextArea();
  JScrollPane jScrollPane2 = new JScrollPane();
  JTextArea resultArea = new JTextArea();
  JButton testFileButton = new JButton();
  JButton exitButton = new JButton();

  /**
   * Contruction method.
   */
  public MainDialog(Frame frame, String title, boolean modal) {
    super(frame, title, modal);
    try {
      jbInit();
      pack();
    }
    catch (Exception ex) {
      ex.printStackTrace();
    }
  }

  public MainDialog() {
    this(null, "A Naive Bayesian Classifier", false);
  }

  private void jbInit() throws Exception {
    ruleArea.setEditable(false);
    ruleArea.setText("");
    resultArea.setEditable(false);
    resultArea.setText("");
    testFileButton.setText("test file...");
    testFileButton.addActionListener(new
                                     MainDialog_testFileButton_actionAdapter(this));
    testButton.addActionListener(new MainDialog_testButton_actionAdapter(this));
    trainButton.addActionListener(new MainDialog_trainButton_actionAdapter(this));
    exitButton.setText("exit");
    exitButton.addActionListener(new MainDialog_exitButton_actionAdapter(this));
    panel1.setLayout(gridBagLayout1);
    panel1.add(jScrollPane2, new GridBagConstraints(0, 4, 4, 2, 0.0, 0.0
        , GridBagConstraints.CENTER, GridBagConstraints.NONE,
        new Insets(10, 2, 11, 0), 550, 100));
    panel1.add(jScrollPane1, new GridBagConstraints(0, 3, 4, 1, 0.0, 0.0
        , GridBagConstraints.SOUTH, GridBagConstraints.NONE,
        new Insets(4, 0, 9, 0), 550, 180));
    panel1.add(browseButton, new GridBagConstraints(4, 0, 1, 1, 0.0, 0.0
        , GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
        new Insets(9, 16, 4, 16), 0, -1));
    panel1.add(jLabel1, new GridBagConstraints(0, 0, 1, 1, 0.0, 0.0
                                               , GridBagConstraints.CENTER,
                                               GridBagConstraints.NONE,
                                               new Insets(0, 9, 0, 0), 7, 0));
    panel1.add(resultClearButton, new GridBagConstraints(4, 5, 1, 1, 0.0, 0.0
        , GridBagConstraints.NORTH, GridBagConstraints.NONE,
        new Insets(8, 3, 80, 6), 20, 0));
    panel1.add(ruleSaveButton, new GridBagConstraints(4, 3, 1, 1, 0.0, 0.0
        , GridBagConstraints.CENTER, GridBagConstraints.NONE,
        new Insets(8, 3, 75, 5), 19, 0));
    panel1.add(fileNameFiled, new GridBagConstraints(1, 0, 3, 1, 0.0, 0.0
        , GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
        new Insets(11, 6, 6, 9), 449, -1));
    panel1.add(trainButton, new GridBagConstraints(1, 1, 1, 1, 0.0, 0.0
        , GridBagConstraints.EAST, GridBagConstraints.NONE,
        new Insets(6, 8, 6, 0), 0, 0));
    panel1.add(testButton, new GridBagConstraints(2, 1, 1, 1, 0.0, 0.0
                                                  , GridBagConstraints.CENTER,
                                                  GridBagConstraints.NONE,
                                                  new Insets(2, 20, 2, 6), 4, 0));
    panel1.add(exitButton, new GridBagConstraints(4, 1, 1, 1, 0.0, 0.0
                                                  , GridBagConstraints.CENTER,
                                                  GridBagConstraints.NONE,
                                                  new Insets(0, 0, 0, 0), 29, 0));
    panel1.add(resultSaveButton, new GridBagConstraints(4, 4, 1, 1, 0.0, 0.0
        , GridBagConstraints.NORTH, GridBagConstraints.NONE,
        new Insets(15, 0, 17, 2), 20, 0));
    panel1.add(testFileButton, new GridBagConstraints(3, 1, 1, 2, 0.0, 0.0
        , GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
        new Insets(6, 14, 6, 2), 8, 0));
    jScrollPane1.getViewport().add(ruleArea, null);
    jScrollPane2.getViewport().add(resultArea, null);
    this.getContentPane().add(panel1, BorderLayout.CENTER);
    ruleSaveButton.setText("save");
    resultClearButton.setText("clear");
    resultSaveButton.setText("save");
    browseButton.setText("browse...");
    browseButton.addActionListener(new MainDialog_browseButton_actionAdapter(this));
    fileNameFiled.setEditable(false);
    fileNameFiled.setText("");
    this.addWindowListener(new MainDialog_this_windowAdapter(this));
    jLabel1.setText("Input file name:");
    testButton.setText("train while testing");
    trainButton.setText("train without testing");
  }

  public static void main(String args[]) {
    MainDialog aDlg = new MainDialog();
    aDlg.setBounds(150, 100, 700, 500);
    aDlg.show();
  }

  /**
   * Handle the event of clicking the "browse..." button.<br>
   * 1. Get the file name <i>XXX</i>;<br>
   * 2. check if the chosen files include <i>XXX.name</i> (required),
   *    <i>XXX.data</i> (required), and <i>XXX.test</i> (optional).<br>
   */
  void browseButton_actionPerformed(ActionEvent e) {
    if (e.getSource() == browseButton) {
      final JFileChooser fc = new JFileChooser();
      fc.setCurrentDirectory(new File("./data"));
      int returnVal = fc.showOpenDialog(this);

      if (returnVal == JFileChooser.APPROVE_OPTION) {
        String pathname = fc.getSelectedFile().getPath();
        String fileitemname = pathname.substring(pathname.lastIndexOf("\\") + 1,
                                                 pathname.lastIndexOf('.'));
        String dirname = pathname.substring(0, pathname.lastIndexOf("\\") + 1);
        // Initial values of these variables are "null".
        this.namefile = new File(dirname + fileitemname + ".names");
        this.datafile = new File(dirname + fileitemname + ".data");
        this.testfile = new File(dirname + fileitemname + ".test");
        if (!namefile.exists()) {
          JOptionPane.showMessageDialog(this, "The .name file doesn't exist!",
                                        "File opening error",
                                        JOptionPane.ERROR_MESSAGE);
          return;
        }
        if (!datafile.exists()) {
          JOptionPane.showMessageDialog(this, "The .data file doesn't exist!",
                                        "File opening error",
                                        JOptionPane.ERROR_MESSAGE);
          return;
        }
        if (!testfile.exists()) {
          this.testButton.setEnabled(false);
          this.testFileButton.setEnabled(false);
        }
        else {
          this.testButton.setEnabled(true);
          this.testFileButton.setEnabled(true);
        }

        this.fileNameFiled.setText(dirname + fileitemname);

        // every time when a new file is specified, RESET the member variables;
        this.trained = false;
        this.tested = false;

        this.classlist = new ArrayList();
        this.attrlist = new ArrayList();
        this.valuelist = new ArrayList();
        //this.valuenumlist = new ArrayList();

        this.tupleNum = 0;

        this.pTable = new Hashtable();
        this.pList = new ArrayList();

        this.confusionMatrix = null;
        this.errorRate = 0;

      }
    }
  }

  /**
   * Handle the event of clicking the "train without testing" button.<br>
   */
  void trainButton_actionPerformed(ActionEvent e) {
    // check if the file name has been specified
    if (this.datafile == null || this.namefile == null) {
      JOptionPane.showMessageDialog(this,
                                    "Please choose a file before training...",
                                    "Internal error", JOptionPane.ERROR_MESSAGE);
      return;
    }

    if (this.trained)
      return; // if the classifier has been trained, simply return

    if (this.loadNameFile() == false)
      return; // read names from the name file
    if (this.train() == false){
      JOptionPane.showMessageDialog(this,
                                    "Training failed !",
                                    "Internal error", JOptionPane.ERROR_MESSAGE);
      return; //// TRAIN the classifier.
    }
    this.trained = true;
    this.tested = false;

    System.out.println("Training complete");

    // print the rules into the "ruleArea".
    StringBuffer buf = new StringBuffer();
    for (int i = 0; i < this.pList.size(); i++) {
      System.out.println(i);
      buf.append((String)this.pList.get(i));
      buf.append("\n");
    }
    this.ruleArea.setText(buf.toString());
    this.resultArea.setText("");
  }

  /**
   * Handle the event of clicking the "train while testing" button.<br>
   */
  void testButton_actionPerformed(ActionEvent e) {
    // check if the file name has been specified
    if (this.datafile == null || this.namefile == null) {
      JOptionPane.showMessageDialog(this,
                                    "Please choose a file before training...",
                                    "Internal error", JOptionPane.ERROR_MESSAGE);
      return;
    }

    if (this.tested)
      return;

    if (!this.trained) {
      if (this.loadNameFile() == false)
        return; // read names from the name file
      if (this.train() == false){
        JOptionPane.showMessageDialog(this,
                                      "Training failed !",
                                      "Internal error", JOptionPane.ERROR_MESSAGE);
        return; //// TRAIN the classifier.
      }
      this.trained = true;

      // print the rules into the "ruleArea".
      String rules = "";
      for (int i = 0; i < this.pList.size(); i++) {
        rules += (String)this.pList.get(i) + "\n";
      }
      this.ruleArea.setText(rules);
    }

    //// TESTING
    if (this.test() == false){
        JOptionPane.showMessageDialog(this,
                                      "Testing failed !",
                                      "Internal error", JOptionPane.ERROR_MESSAGE);
        return; //// TRAIN the classifier.
      }
    this.tested = true;

    // print the testing results into the "resultArea"
    String result = "total # of testing cases: " + tupleNum;
    result += "  (average error rate: " +
        this.double2string(this.errorRate * 100, 2) + "%)\n\n";
    result += "   ----------------Confusion Matrix----------------\n    ";
    for (int i = 0; i < this.classlist.size(); i++) {
      result += (String)this.classlist.get(i) + "\t";
    }
    result += "<-- predicted classes\n";

    for (int i = 0; i < this.classlist.size(); i++) {
      result += "    ";
      for (int j = 0; j < this.classlist.size(); j++) {
        int val = this.confusionMatrix[i][j];
        result += val + "\t";
      }
      result += "<-- actual class: '" + (String)this.classlist.get(i) + "'\n";
    }
    this.resultArea.setText(result);
  }

  /**
   * Handle the event of clicking the "test file..." button.<br>
   */
  void testFileButton_actionPerformed(ActionEvent e) {
    if (this.testfile == null) {
      return;
    }
  }

  /**
   * Read the class names and the attribute names from the name file to the "classlist" and "attrlist".<br>
   */
  protected boolean loadNameFile() {
    try {
      BufferedReader reader = new BufferedReader(new FileReader(this.namefile));
      String line = "";
      int ind = 0;

      boolean inAttrScope = false;
      while ( (line = reader.readLine()) != null) {
        // remove the annotation
        ind = line.indexOf('|');
        if (ind >= 0) {
          line = line.trim().substring(0, ind).trim();
        }
        if (line.equals(""))
          continue;

        ind = line.indexOf(':');

        if (ind >= 0) {
          inAttrScope = true;
          // read attribute names into the "attrlist".
          String name = line.substring(0, ind).trim();
          this.attrlist.add(name);
          this.valuelist.add(new ArrayList()); // for containing the value list of the attribute.
        }
        else {
          if(inAttrScope) continue;
          // read class names (in a single line) into "classlist"
          ind = line.indexOf(',');
          while (ind > 0) {
            String cls = line.substring(0, ind).trim();
            classlist.add(cls);
            line = line.substring(ind + 1).trim();
            ind = line.indexOf(',');
          }
          if(!line.equals("")){
            if(line.charAt(line.length()-1) == '.')
              line = line.substring(0, line.length()-1);
              classlist.add(line);
          }
        }
      } // while readLine()

      return true;

    }
    catch (IOException ex) {
      JOptionPane.showMessageDialog(this,
          "File I/O error. Please verify the name file",
                                    "Internal error", JOptionPane.ERROR_MESSAGE);
      ex.printStackTrace();
      return false;
    }

  }

  /**
   * Read data from the data file and train the classifier
   */
  protected boolean train() {
    try {
      BufferedReader reader = new BufferedReader(new FileReader(this.datafile));
      String line = "";
      int ind = 0;
      this.tupleNum = 0;
      while ( (line = reader.readLine()) != null) {
        // remove the annotation
        ind = line.indexOf('|');
        if (ind >= 0) {
          line = line.trim().substring(0, ind).trim();
        }
        if (line.equals(""))
          continue;

        this.tupleNum++;

        // get the class name firstly
        String clsname = "";
        ind = line.lastIndexOf(',');
        if (ind > 0) {
          if(line.lastIndexOf('.') > ind)
            clsname = line.substring(ind + 1, line.lastIndexOf('.')).trim();
          else
            clsname = line.substring(ind + 1).trim();
          line = line.substring(0, ind + 1).trim(); // keep the last ','
        }
        else
          return false;

        int index1 = -1; // index in "attrlist"
        ind = line.indexOf(',');
        while (ind > 0) {
          index1++;
          String attrvalue = line.substring(0, ind);

          line = line.substring(ind + 1).trim();
          ind = line.indexOf(',');
          if (attrvalue.equals("?"))
            continue;

          //// add different values of the current attribute into the "valuelist".
          //// count the number of a value of the current attribute, into the "valuenumlist"
          ArrayList values = (ArrayList)this.valuelist.get(index1);
          if (!values.contains(attrvalue)) {
            values.add(attrvalue);
          }
          this.valuelist.set(index1, values);

          //// Operations on the hashtable
          // 1. count the attribute: count of "P(Attr1=a1|Class=c1)"
          String attrname = (String)this.attrlist.get(index1);
          String key = "P(" + attrname + '=' + attrvalue + "|class=" + clsname +
              ")";
          int oldNum = 0;
          if (this.pTable.containsKey(key)) {
            oldNum = ( (Integer) pTable.get(key)).intValue();
            this.pTable.remove(key);
          }
          this.pTable.put(key, new Integer(oldNum + 1));
        }

        // 2. count the class: count of "P(Class=c1)"
        String key = "P(class=" + clsname + ")";
        int oldNum = 0;
        if (this.pTable.containsKey(key)) {
          oldNum = ( (Integer) pTable.get(key)).intValue();
          this.pTable.remove(key);
        }
        this.pTable.put(key, new Integer(oldNum + 1));

        System.out.println("Reading the "+this.tupleNum+" tuple");
      } // while readLine()

      //// Compute the probabilities
      double p = 0;
      String key = "";
      for (int i = 0; i < this.classlist.size(); i++) {
        String clsname = (String)this.classlist.get(i);
        // 1. compute "P(Class=c1)"
        key = "P(class=" + clsname + ")";

        System.out.println("Computing "+key);

        double ni;
        if (this.pTable.containsKey(key))
          ni = (double) ( (Integer)this.pTable.get(key)).intValue();
        else
          ni = 0;
        p = (double) (ni / ( (double)this.tupleNum));
        this.pTable.remove(key);
        this.pTable.put(key, new Double(p));
        this.pList.add(key + "=" + this.double2string(p, 2)); // add into the "pList" the prob of the class

        // 2. compute "P(Attr1=a1|Class=c1)"
        for (int j = 0; j < this.attrlist.size(); j++) {
          String attrname = (String)this.attrlist.get(j);
          ArrayList values = (ArrayList)this.valuelist.get(j);

          double nj = (double) values.size();

          for (int m = 0; m < values.size(); m++) {
            String attrvalue = (String) values.get(m);
            key = "P(" + attrname + '=' + attrvalue + "|class=" + clsname + ")";

            System.out.println("Computing "+key);

            // the computation needs to be smoothed. !!! SMOOTHING !!!
            double nijk;
            if (this.pTable.containsKey(key))
              nijk = (double) ( (Integer)this.pTable.get(key)).intValue();
            else
              nijk = 0;
            double f = (double) (1 / (double)this.tupleNum);
            p = (double) ( (nijk + f) / (ni + f * nj));
            this.pTable.remove(key);
            this.pTable.put(key, new Double(p));
            this.pList.add(key + "=" + this.double2string(p, 2)); // add into the "pList" the prob of the class
          }
        }
      }

      return true;

    }
    catch (IOException ex) {
      JOptionPane.showMessageDialog(this,
          "File I/O error. Please verify the data file",
                                    "Internal error", JOptionPane.ERROR_MESSAGE);
      ex.printStackTrace();
      return false;
    }

  }

  /**
   * Read data from the test file and do testing on the data
   */
  protected boolean test() {
    try {
      // initialize the confusion matrix. Row is the actual class, Column is the predicted class
      this.confusionMatrix = new int[this.classlist.size()][this.classlist.size()];
      for (int i = 0; i < this.classlist.size(); i++)
        for (int j = 0; j < this.classlist.size(); j++)
          this.confusionMatrix[i][j] = 0;

      BufferedReader reader = new BufferedReader(new FileReader(this.testfile));
      String line = "";
      int ind = 0;
      this.tupleNum = 0;
      int row = 0; // used to index the row of the matrix
      int col = 0; // used to index the col of the matrix
      while ( (line = reader.readLine()) != null) {
        // remove the annotation
        ind = line.indexOf('|');
        if (ind >= 0) {
          line = line.trim().substring(0, ind).trim();
        }
        if (line.equals(""))
          continue;

        this.tupleNum++;

        // get the ACTUAL class name firstly
        String actclsname = "";
        ind = line.lastIndexOf(',');
        if (ind > 0) {
          if(line.lastIndexOf('.') > ind)
            actclsname = line.substring(ind + 1, line.lastIndexOf('.')).trim();
          else
            actclsname = line.substring(ind + 1).trim();
          line = line.substring(0, ind + 1).trim(); // keep the last ','
          row = this.classlist.indexOf(actclsname);
        }
        else
          return false;

        // for storing the prob. for "P(A1=a1^...^An=an | C=ci)", where i = 1,2,...,m
        ArrayList probList = new ArrayList();
        for (int i = 0; i < this.classlist.size(); i++) {
          String key = "P(class=" + (String)this.classlist.get(i) + ")";
          if (!this.pTable.containsKey(key))
            return false;
          probList.add( (Double)this.pTable.get(key));
        }

        int index1 = -1; // index in "attrlist"
        ind = line.indexOf(',');
        while (ind > 0) {
          index1++;
          String attrvalue = line.substring(0, ind);

          line = line.substring(ind + 1).trim();
          ind = line.indexOf(',');
          if (attrvalue.equals("?"))
            continue;

          // 1. compute the "P(A1=a1^...^An=an | C=ci)", where i = 1,2,...,m
          String attrname = (String)this.attrlist.get(index1);
          for (int i = 0; i < this.classlist.size(); i++) {
            String key = "P(" + attrname + '=' + attrvalue + "|class=" +
                (String)this.classlist.get(i) + ")";
            if (!this.pTable.containsKey(key))
              return false;
            double old_p = ( (Double) probList.get(i)).doubleValue();
            double cur_p = ( (Double)this.pTable.get(key)).doubleValue();
            probList.set(i, new Double(old_p * cur_p));
          }
        }

        // 2. decide what class the current tuple belongs to...
        double maxProb = 0;
        int maxIndex = 0;
        for (int i = 0; i < probList.size(); i++) {
          double curProb = ( (Double) probList.get(i)).doubleValue();
          if (curProb > maxProb) {
            maxProb = curProb;
            maxIndex = i;
          }
        }

        col = maxIndex;
        this.confusionMatrix[row][col]++;
      } // while readLine()

      // 3. compute the error rate
      int errorNum = 0;
      for (int i = 0; i < this.classlist.size(); i++)
        for (int j = 0; j < this.classlist.size(); j++)
          if (i != j)
            errorNum += this.confusionMatrix[i][j];
      this.errorRate = (double) errorNum / (double) tupleNum;

      return true;

    }
    catch (IOException ex) {
      JOptionPane.showMessageDialog(this,
          "File I/O error. Please verify the test file",
                                    "Internal error", JOptionPane.ERROR_MESSAGE);
      ex.printStackTrace();
      return false;
    }

  }

  private String double2string(double d, int decimal) {
    String str = Double.toString(d);
    int length = str.indexOf('.') + decimal + 1;
    if (str.length() < length)
      length = str.length();
    return str.substring(0, length);
  }

  void this_windowClosing(WindowEvent e) {
    System.exit(1);
  }

  void exitButton_actionPerformed(ActionEvent e) {
    System.exit(1);
  }

}

class MainDialog_browseButton_actionAdapter
    implements java.awt.event.ActionListener {
  MainDialog adaptee;

  MainDialog_browseButton_actionAdapter(MainDialog adaptee) {
    this.adaptee = adaptee;
  }

  public void actionPerformed(ActionEvent e) {
    adaptee.browseButton_actionPerformed(e);
  }
}

class MainDialog_this_windowAdapter
    extends java.awt.event.WindowAdapter {
  MainDialog adaptee;

  MainDialog_this_windowAdapter(MainDialog adaptee) {
    this.adaptee = adaptee;
  }

  public void windowClosing(WindowEvent e) {
    adaptee.this_windowClosing(e);
  }
}

class MainDialog_testButton_actionAdapter
    implements java.awt.event.ActionListener {
  MainDialog adaptee;

  MainDialog_testButton_actionAdapter(MainDialog adaptee) {
    this.adaptee = adaptee;
  }

  public void actionPerformed(ActionEvent e) {
    adaptee.testButton_actionPerformed(e);
  }
}

class MainDialog_trainButton_actionAdapter
    implements java.awt.event.ActionListener {
  MainDialog adaptee;

  MainDialog_trainButton_actionAdapter(MainDialog adaptee) {
    this.adaptee = adaptee;
  }

  public void actionPerformed(ActionEvent e) {
    adaptee.trainButton_actionPerformed(e);
  }
}

class MainDialog_testFileButton_actionAdapter
    implements java.awt.event.ActionListener {
  MainDialog adaptee;

  MainDialog_testFileButton_actionAdapter(MainDialog adaptee) {
    this.adaptee = adaptee;
  }

  public void actionPerformed(ActionEvent e) {
    adaptee.testFileButton_actionPerformed(e);
  }
}

class MainDialog_exitButton_actionAdapter
    implements java.awt.event.ActionListener {
  MainDialog adaptee;

  MainDialog_exitButton_actionAdapter(MainDialog adaptee) {
    this.adaptee = adaptee;
  }

  public void actionPerformed(ActionEvent e) {
    adaptee.exitButton_actionPerformed(e);
  }
}